
import logging
from typing import (
    TYPE_CHECKING,
    Tuple,
)

from opentelemetry.trace import (
    SpanKind,
    StatusCode,
    TraceFlags,
)
from opentelemetry.sdk.trace import SpanProcessor
from opentelemetry.semconv.trace import SpanAttributes

if TYPE_CHECKING:
    from opentelemetry.sdk.trace import ReadableSpan
    from solarwinds_apm.apm_txname_manager import SolarWindsTxnNameManager


logger = logging.getLogger(__name__)


class SolarWindsInboundMetricsSpanProcessor(SpanProcessor):

    _HTTP_METHOD = SpanAttributes.HTTP_METHOD            # "http.method"
    _HTTP_ROUTE = SpanAttributes.HTTP_ROUTE              # "http.route"
    _HTTP_STATUS_CODE = SpanAttributes.HTTP_STATUS_CODE  # "http.status_code"
    _HTTP_URL = SpanAttributes.HTTP_URL                  # "http.url"

    _LIBOBOE_HTTP_SPAN_STATUS_UNAVAILABLE = 0

    def __init__(
        self,
        apm_txname_manager: "SolarWindsTxnNameManager",
        agent_enabled: bool,
    ) -> None:
        self._apm_txname_manager = apm_txname_manager
        if agent_enabled:
            from solarwinds_apm.extension.oboe import Span
            self._span = Span
        else:
            from solarwinds_apm.apm_noop import Span
            self._span = Span

    def on_end(self, span: "ReadableSpan") -> None:
        """Calculates and reports inbound trace metrics,
        and caches liboboe transaction name"""
        # Only calculate inbound metrics for service root spans
        parent_span_context = span.parent
        if parent_span_context and parent_span_context.is_valid \
            and not parent_span_context.is_remote:
            return

        is_span_http = self.is_span_http(span)
        span_time = self.calculate_span_time(
            span.start_time,
            span.end_time,
        )
        # TODO Use `domain` for custom transaction naming after alpha/beta
        domain = None
        has_error = self.has_error(span)
        trans_name, url_tran = self.calculate_transaction_names(span)

        liboboe_txn_name = None
        if is_span_http:
            # createHttpSpan needs these other params
            status_code = self.get_http_status_code(span)
            request_method = span.attributes.get(self._HTTP_METHOD, None)

            logger.debug(
                "createHttpSpan with trans_name: {}, url_tran: {}, domain: {}, span_time: {} status_code: {}, request_method: {}, has_error: {}".format(
                    trans_name, url_tran, domain, span_time, status_code, request_method, has_error,
                )
            )
            liboboe_txn_name = self._span.createHttpSpan(
                trans_name,
                url_tran,
                domain,
                span_time,
                status_code,
                request_method,
                has_error,
            )  
        else:
            logger.debug(
                "createSpan with trans_name: {}, domain: {}, span_time: {}, has_error: {}".format(
                    trans_name, domain, span_time, has_error,
                )
            )
            liboboe_txn_name = self._span.createSpan(
                trans_name,
                domain,
                span_time,
                has_error,
            )

        if span.context.trace_flags == TraceFlags.SAMPLED:
            # Cache txn_name for span export
            self._apm_txname_manager[
                "{}-{}".format(span.context.trace_id, span.context.span_id)
            ] = liboboe_txn_name

    def is_span_http(self, span: "ReadableSpan") -> bool:
        """This span from inbound HTTP request if from a SERVER by some http.method"""
        if span.kind == SpanKind.SERVER and span.attributes.get(self._HTTP_METHOD, None):
            return True
        return False

    def has_error(self, span: "ReadableSpan") -> bool:
        """Calculate if this span instance has_error"""
        if span.status.status_code == StatusCode.ERROR:
            return True
        return False

    def get_http_status_code(self, span: "ReadableSpan") -> int:
        """Calculate HTTP status_code from span or default to UNAVAILABLE"""
        status_code = span.attributes.get(self._HTTP_STATUS_CODE, None)
        # Something went wrong in OTel or instrumented service crashed early
        # if no status_code in attributes of HTTP span
        if not status_code:
            status_code = self._LIBOBOE_HTTP_SPAN_STATUS_UNAVAILABLE
        return status_code

    def calculate_transaction_names(self, span: "ReadableSpan") -> Tuple[str, str]:
        """Get trans_name and url_tran of this span instance."""
        url_tran = span.attributes.get(self._HTTP_URL, None)
        http_route = span.attributes.get(self._HTTP_ROUTE, None)
        trans_name = None
        if http_route:
            trans_name = http_route
        elif span.name:
            trans_name = span.name
        return trans_name, url_tran

    def calculate_span_time(
        self,
        start_time: int,
        end_time: int,
    ) -> int:
        """Calculate span time in microseconds (us) using start and end time
        in nanoseconds (ns). OTel span start/end_time are optional."""
        if not start_time or not end_time:
            return 0
        return int((end_time - start_time) // 1e3)



import logging
import typing

from opentelemetry import trace
from opentelemetry.context.context import Context
from opentelemetry.instrumentation.propagators import ResponsePropagator
from opentelemetry.propagators import textmap
from opentelemetry.trace.span import TraceState

from solarwinds_apm.apm_constants import (
    INTL_SWO_COMMA,
    INTL_SWO_COMMA_W3C_SANITIZED,
    INTL_SWO_EQUALS,
    INTL_SWO_EQUALS_W3C_SANITIZED
)
from solarwinds_apm.traceoptions import XTraceOptions
from solarwinds_apm.w3c_transformer import W3CTransformer

logger = logging.getLogger(__name__)

class SolarWindsTraceResponsePropagator(ResponsePropagator):
    """Propagator that injects SW values into HTTP responses"""
    _HTTP_HEADER_ACCESS_CONTROL_EXPOSE_HEADERS = "Access-Control-Expose-Headers"
    _XTRACE_HEADER_NAME = "x-trace"
    _XTRACEOPTIONS_RESPONSE_HEADER_NAME = "x-trace-options-response"
    
    def inject(
        self,
        carrier: textmap.CarrierT,
        context: typing.Optional[Context] = None,
        setter: textmap.Setter = textmap.default_setter,
    ) -> None:
        """Injects x-trace and options-response into the HTTP response carrier."""
        span = trace.get_current_span(context)
        span_context = span.get_span_context()
        if span_context == trace.INVALID_SPAN_CONTEXT:
            return
        
        x_trace = W3CTransformer.traceparent_from_context(span_context)
        setter.set(
            carrier,
            self._XTRACE_HEADER_NAME,
            x_trace,
        )
        exposed_headers = [self._XTRACE_HEADER_NAME]

        xtraceoptions_response = self.recover_response_from_tracestate(
            span_context.trace_state
        )
        if xtraceoptions_response:
            exposed_headers.append("{}".format(
                self._XTRACEOPTIONS_RESPONSE_HEADER_NAME
            ))
            setter.set(
                carrier,
                self._XTRACEOPTIONS_RESPONSE_HEADER_NAME,
                xtraceoptions_response,
            )
        setter.set(
            carrier,
            self._HTTP_HEADER_ACCESS_CONTROL_EXPOSE_HEADERS,
            ",".join(exposed_headers),
        )
    
    def recover_response_from_tracestate(
        self,
        tracestate: TraceState
    ) -> str:
        """Use tracestate to recover xtraceoptions response by
        converting delimiters:
        EQUALS_W3C_SANITIZED becomes EQUALS
        COMMA_W3C_SANITIZED becomes COMMA
        """
        sanitized = tracestate.get(
            XTraceOptions.get_sw_xtraceoptions_response_key(),
            None
        )
        if not sanitized:
            return
        return sanitized.replace(
            INTL_SWO_EQUALS_W3C_SANITIZED,
            INTL_SWO_EQUALS
        ).replace(INTL_SWO_COMMA_W3C_SANITIZED, INTL_SWO_COMMA)



"""Module to configure OpenTelemetry to work with SolarWinds backend"""

from os import environ

from opentelemetry.environment_variables import (
    OTEL_PROPAGATORS,
    OTEL_TRACES_EXPORTER
)
from opentelemetry.instrumentation.distro import BaseDistro

from solarwinds_apm.apm_constants import (
    INTL_SWO_DEFAULT_TRACES_EXPORTER,
    INTL_SWO_DEFAULT_PROPAGATORS
) 


class SolarWindsDistro(BaseDistro):
    """OpenTelemetry Distro for SolarWinds reporting environment"""

    def _configure(self, **kwargs):
        """Configure default OTel exporter and propagators"""
        environ.setdefault(OTEL_TRACES_EXPORTER, INTL_SWO_DEFAULT_TRACES_EXPORTER)
        environ.setdefault(OTEL_PROPAGATORS, ",".join(INTL_SWO_DEFAULT_PROPAGATORS))



#    Copyright 2022 SolarWinds, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" SolarWinds instrumentation API for Python.

apm_noop defines no-op classes for platforms we don't support building the c extension on
"""
# No-op classes intentionally left undocumented
# "Missing docstring"
# pylint: disable-msg=C0103
import threading


class Metadata:
    def __init__(self, _=None):
        pass

    @staticmethod
    def fromString(_):
        return Metadata()

    def createEvent(self):
        return Event()

    @staticmethod
    def makeRandom(flag=True):
        return Metadata(True)

    def copy(self):
        return self

    def isValid(self):
        return False

    def isSampled(self):
        return False

    def toString(self):
        return ''


class Context:
    transaction_dict = threading.local()

    @staticmethod
    def setTracingMode(_):
        return False

    @staticmethod
    def setTriggerMode(_):
        return False

    @staticmethod
    def setDefaultSampleRate(_):
        return False

    @staticmethod
    def getDecisions(*args, **kwargs):
        """
        This is the stub method in no-op mode. The corresponding method in the
        `normal` mode (when the instrumentation is working) is actually a wrapper
        of the `getDecisions` method of the Context class in swig/oboe.hpp.

        There are multiple return values in this method, which are:
        do_metrics, do_sample, sample_rate, sample_source, bucket_rate, bucket_capacity, typ, auth, status_msg,
        auth_msg, status

        Go to oboe_api.h to see the descriptions of all the return values.
        """
        return 0, 0, 0, 6, 0, 0, 0, -1, "", "", 0

    @staticmethod
    def get():
        return Metadata()

    @staticmethod
    def set(_):
        pass

    @staticmethod
    def fromString(_):
        return Context()

    @staticmethod
    def copy():
        return Context()

    @staticmethod
    def clear():
        pass

    @staticmethod
    def isValid():
        return False

    @staticmethod
    def toString():
        return ''

    @staticmethod
    def createEvent():
        return Event()

    @staticmethod
    def startTrace(_=None):
        return Event()

    @staticmethod
    def isReady(_):
        # unknown error
        return 0


class Event:
    def __init__(self, _=None, __=None):
        pass

    def addInfo(self, _, __):
        pass

    def addEdge(self, _):
        pass

    def addEdgeStr(self, _):
        pass

    def getMetadata(self):
        return Metadata()

    def metadataString(self):
        return ''

    def is_valid(self):
        return False

    @staticmethod
    def startTrace(_=None):
        return Event()


class Reporter:
    def __init__(self, *args, **kwargs):
        self.init_status = 0

    def sendReport(self, _, __=None):
        pass

    def sendStatus(self, _, __=None):
        pass

    def flush(self):
        pass


class Span:
    @staticmethod
    def createHttpSpan(_, __, ___, ____, _____, ______, _______):
        pass

    @staticmethod
    def createSpan(_, __, ___, ____):
        pass


class MetricTags:
    def __init__(self, count):
        super(MetricTags, self).__init__()

    @staticmethod
    def add(*args, **kwargs):
        pass


class CustomMetrics:
    @staticmethod
    def summary(*args, **kwargs):
        pass

    @staticmethod
    def increment(*args, **kwargs):
        pass


class Config:
    @staticmethod
    def getVersionString():
        return "No extension loaded."



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 4.0.2
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
if _swig_python_version_info < (2, 7, 0):
    raise RuntimeError("Python 2.7 or later required")

# Import the low-level C/C++ module
if __package__ or "." in __name__:
    from . import _oboe
else:
    import _oboe

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)


def _swig_setattr_nondynamic_instance_variable(set):
    def set_instance_attr(self, name, value):
        if name == "thisown":
            self.this.own(value)
        elif name == "this":
            set(self, name, value)
        elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
            set(self, name, value)
        else:
            raise AttributeError("You cannot add instance attributes to %s" % self)
    return set_instance_attr


def _swig_setattr_nondynamic_class_variable(set):
    def set_class_attr(cls, name, value):
        if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
            set(cls, name, value)
        else:
            raise AttributeError("You cannot add class attributes to %s" % cls)
    return set_class_attr


def _swig_add_metaclass(metaclass):
    """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
    def wrapper(cls):
        return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
    return wrapper


class _SwigNonDynamicMeta(type):
    """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
    __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)


OBOE_DEBUG_DISABLED = _oboe.OBOE_DEBUG_DISABLED
OBOE_DEBUG_FATAL = _oboe.OBOE_DEBUG_FATAL
OBOE_DEBUG_ERROR = _oboe.OBOE_DEBUG_ERROR
OBOE_DEBUG_WARNING = _oboe.OBOE_DEBUG_WARNING
OBOE_DEBUG_INFO = _oboe.OBOE_DEBUG_INFO
OBOE_DEBUG_LOW = _oboe.OBOE_DEBUG_LOW
OBOE_DEBUG_MEDIUM = _oboe.OBOE_DEBUG_MEDIUM
OBOE_DEBUG_HIGH = _oboe.OBOE_DEBUG_HIGH
OBOE_MODULE_ALL = _oboe.OBOE_MODULE_ALL
OBOE_MODULE_UNDEF = _oboe.OBOE_MODULE_UNDEF
OBOE_MODULE_LIBOBOE = _oboe.OBOE_MODULE_LIBOBOE
OBOE_MODULE_SETTINGS = _oboe.OBOE_MODULE_SETTINGS
OBOE_MODULE_REPORTER_FILE = _oboe.OBOE_MODULE_REPORTER_FILE
OBOE_MODULE_REPORTER_UDP = _oboe.OBOE_MODULE_REPORTER_UDP
OBOE_MODULE_REPORTER_SSL = _oboe.OBOE_MODULE_REPORTER_SSL
OBOE_MODULE_REPORTER_LAMBDA = _oboe.OBOE_MODULE_REPORTER_LAMBDA
OBOE_MODULE_APACHE = _oboe.OBOE_MODULE_APACHE
OBOE_MODULE_NGINX = _oboe.OBOE_MODULE_NGINX
OBOE_MODULE_PHP = _oboe.OBOE_MODULE_PHP
OBOE_MODULE_DOTNET = _oboe.OBOE_MODULE_DOTNET
OBOE_MODULE_RUBY = _oboe.OBOE_MODULE_RUBY
class Metadata(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, md):
        _oboe.Metadata_swiginit(self, _oboe.new_Metadata(md))
    __swig_destroy__ = _oboe.delete_Metadata

    def createEvent(self):
        return _oboe.Metadata_createEvent(self)

    def copy(self):
        return _oboe.Metadata_copy(self)

    def isValid(self):
        return _oboe.Metadata_isValid(self)

    def isSampled(self):
        return _oboe.Metadata_isSampled(self)

    @staticmethod
    def makeRandom(sampled=True):
        return _oboe.Metadata_makeRandom(sampled)

    @staticmethod
    def fromString(s):
        return _oboe.Metadata_fromString(s)

    def metadata(self):
        return _oboe.Metadata_metadata(self)

    def toString(self):
        return _oboe.Metadata_toString(self)

# Register Metadata in _oboe:
_oboe.Metadata_swigregister(Metadata)

def Metadata_makeRandom(sampled=True):
    return _oboe.Metadata_makeRandom(sampled)

def Metadata_fromString(s):
    return _oboe.Metadata_fromString(s)

class Context(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    @staticmethod
    def setTracingMode(newMode):
        return _oboe.Context_setTracingMode(newMode)

    @staticmethod
    def setTriggerMode(newMode):
        return _oboe.Context_setTriggerMode(newMode)

    @staticmethod
    def setDefaultSampleRate(newRate):
        return _oboe.Context_setDefaultSampleRate(newRate)

    @staticmethod
    def getDecisions(*args):
        return _oboe.Context_getDecisions(*args)

    @staticmethod
    def get():
        return _oboe.Context_get()

    @staticmethod
    def toString():
        return _oboe.Context_toString()

    @staticmethod
    def set(md):
        return _oboe.Context_set(md)

    @staticmethod
    def fromString(s):
        return _oboe.Context_fromString(s)

    @staticmethod
    def copy():
        return _oboe.Context_copy()

    @staticmethod
    def setSampledFlag():
        return _oboe.Context_setSampledFlag()

    @staticmethod
    def clear():
        return _oboe.Context_clear()

    @staticmethod
    def isValid():
        return _oboe.Context_isValid()

    @staticmethod
    def isSampled():
        return _oboe.Context_isSampled()

    @staticmethod
    def validateTransformServiceName(service_key):
        return _oboe.Context_validateTransformServiceName(service_key)

    @staticmethod
    def shutdown():
        return _oboe.Context_shutdown()

    @staticmethod
    def isReady(timeout):
        return _oboe.Context_isReady(timeout)

    @staticmethod
    def isLambda():
        return _oboe.Context_isLambda()

    @staticmethod
    def startTrace():
        return _oboe.Context_startTrace()

    @staticmethod
    def createEntry(md, timestamp, parent_md=None):
        return _oboe.Context_createEntry(md, timestamp, parent_md)

    @staticmethod
    def createEvent(*args):
        return _oboe.Context_createEvent(*args)

    @staticmethod
    def createExit(timestamp):
        return _oboe.Context_createExit(timestamp)

    def __init__(self):
        _oboe.Context_swiginit(self, _oboe.new_Context())
    __swig_destroy__ = _oboe.delete_Context

# Register Context in _oboe:
_oboe.Context_swigregister(Context)

def Context_setTracingMode(newMode):
    return _oboe.Context_setTracingMode(newMode)

def Context_setTriggerMode(newMode):
    return _oboe.Context_setTriggerMode(newMode)

def Context_setDefaultSampleRate(newRate):
    return _oboe.Context_setDefaultSampleRate(newRate)

def Context_getDecisions(*args):
    return _oboe.Context_getDecisions(*args)

def Context_get():
    return _oboe.Context_get()

def Context_toString():
    return _oboe.Context_toString()

def Context_set(md):
    return _oboe.Context_set(md)

def Context_fromString(s):
    return _oboe.Context_fromString(s)

def Context_copy():
    return _oboe.Context_copy()

def Context_setSampledFlag():
    return _oboe.Context_setSampledFlag()

def Context_clear():
    return _oboe.Context_clear()

def Context_isValid():
    return _oboe.Context_isValid()

def Context_isSampled():
    return _oboe.Context_isSampled()

def Context_validateTransformServiceName(service_key):
    return _oboe.Context_validateTransformServiceName(service_key)

def Context_shutdown():
    return _oboe.Context_shutdown()

def Context_isReady(timeout):
    return _oboe.Context_isReady(timeout)

def Context_isLambda():
    return _oboe.Context_isLambda()

def Context_startTrace():
    return _oboe.Context_startTrace()

def Context_createEntry(md, timestamp, parent_md=None):
    return _oboe.Context_createEntry(md, timestamp, parent_md)

def Context_createEvent(*args):
    return _oboe.Context_createEvent(*args)

def Context_createExit(timestamp):
    return _oboe.Context_createExit(timestamp)

class Event(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined")
    __repr__ = _swig_repr
    __swig_destroy__ = _oboe.delete_Event

    def addInfo(self, *args):
        return _oboe.Event_addInfo(self, *args)

    def addEdge(self, md):
        return _oboe.Event_addEdge(self, md)

    def addContextOpId(self, md):
        return _oboe.Event_addContextOpId(self, md)

    def addHostname(self):
        return _oboe.Event_addHostname(self)

    def getMetadata(self):
        return _oboe.Event_getMetadata(self)

    def metadataString(self):
        return _oboe.Event_metadataString(self)

    def opIdString(self):
        return _oboe.Event_opIdString(self)

    def send(self, with_system_timestamp=True):
        return _oboe.Event_send(self, with_system_timestamp)

    def sendProfiling(self):
        return _oboe.Event_sendProfiling(self)

    def addSpanRef(self, md):
        return _oboe.Event_addSpanRef(self, md)

    def addProfileEdge(self, id):
        return _oboe.Event_addProfileEdge(self, id)

    @staticmethod
    def startTrace(md):
        return _oboe.Event_startTrace(md)

# Register Event in _oboe:
_oboe.Event_swigregister(Event)

def Event_startTrace(md):
    return _oboe.Event_startTrace(md)

class Span(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    @staticmethod
    def createSpan(transaction, domain, duration, has_error, service_name=None):
        return _oboe.Span_createSpan(transaction, domain, duration, has_error, service_name)

    @staticmethod
    def createHttpSpan(transaction, url, domain, duration, status, method, has_error, service_name=None):
        return _oboe.Span_createHttpSpan(transaction, url, domain, duration, status, method, has_error, service_name)

    def __init__(self):
        _oboe.Span_swiginit(self, _oboe.new_Span())
    __swig_destroy__ = _oboe.delete_Span

# Register Span in _oboe:
_oboe.Span_swigregister(Span)

def Span_createSpan(transaction, domain, duration, has_error, service_name=None):
    return _oboe.Span_createSpan(transaction, domain, duration, has_error, service_name)

def Span_createHttpSpan(transaction, url, domain, duration, status, method, has_error, service_name=None):
    return _oboe.Span_createHttpSpan(transaction, url, domain, duration, status, method, has_error, service_name)

class MetricTags(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, count):
        _oboe.MetricTags_swiginit(self, _oboe.new_MetricTags(count))
    __swig_destroy__ = _oboe.delete_MetricTags

    def add(self, index, k, v):
        return _oboe.MetricTags_add(self, index, k, v)

    def get(self):
        return _oboe.MetricTags_get(self)

# Register MetricTags in _oboe:
_oboe.MetricTags_swigregister(MetricTags)

class CustomMetrics(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    @staticmethod
    def summary(name, value, count, host_tag, service_name, tags, tags_count):
        return _oboe.CustomMetrics_summary(name, value, count, host_tag, service_name, tags, tags_count)

    @staticmethod
    def increment(name, count, host_tag, service_name, tags, tags_count):
        return _oboe.CustomMetrics_increment(name, count, host_tag, service_name, tags, tags_count)

    def __init__(self):
        _oboe.CustomMetrics_swiginit(self, _oboe.new_CustomMetrics())
    __swig_destroy__ = _oboe.delete_CustomMetrics

# Register CustomMetrics in _oboe:
_oboe.CustomMetrics_swigregister(CustomMetrics)

def CustomMetrics_summary(name, value, count, host_tag, service_name, tags, tags_count):
    return _oboe.CustomMetrics_summary(name, value, count, host_tag, service_name, tags, tags_count)

def CustomMetrics_increment(name, count, host_tag, service_name, tags, tags_count):
    return _oboe.CustomMetrics_increment(name, count, host_tag, service_name, tags, tags_count)

class Reporter(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    init_status = property(_oboe.Reporter_init_status_get, _oboe.Reporter_init_status_set)

    def __init__(self, hostname_alias, log_level, log_file_path, max_transactions, max_flush_wait_time, events_flush_interval, max_request_size_bytes, reporter, host, service_key, trusted_path, buffer_size, trace_metrics, histogram_precision, token_bucket_capacity, token_bucket_rate, file_single, ec2_metadata_timeout, grpc_proxy, stdout_clear_nonblocking, is_grpc_clean_hack_enabled, w3c_trace_format, metric_format):
        _oboe.Reporter_swiginit(self, _oboe.new_Reporter(hostname_alias, log_level, log_file_path, max_transactions, max_flush_wait_time, events_flush_interval, max_request_size_bytes, reporter, host, service_key, trusted_path, buffer_size, trace_metrics, histogram_precision, token_bucket_capacity, token_bucket_rate, file_single, ec2_metadata_timeout, grpc_proxy, stdout_clear_nonblocking, is_grpc_clean_hack_enabled, w3c_trace_format, metric_format))
    __swig_destroy__ = _oboe.delete_Reporter

    def sendReport(self, *args):
        return _oboe.Reporter_sendReport(self, *args)

    def sendStatus(self, *args):
        return _oboe.Reporter_sendStatus(self, *args)

    def flush(self):
        return _oboe.Reporter_flush(self)

    def getType(self):
        return _oboe.Reporter_getType(self)

# Register Reporter in _oboe:
_oboe.Reporter_swigregister(Reporter)

class Config(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    @staticmethod
    def checkVersion(version, revision):
        return _oboe.Config_checkVersion(version, revision)

    @staticmethod
    def getVersionString():
        return _oboe.Config_getVersionString()

    def __init__(self):
        _oboe.Config_swiginit(self, _oboe.new_Config())
    __swig_destroy__ = _oboe.delete_Config

# Register Config in _oboe:
_oboe.Config_swigregister(Config)

def Config_checkVersion(version, revision):
    return _oboe.Config_checkVersion(version, revision)

def Config_getVersionString():
    return _oboe.Config_getVersionString()







import logging
from typing import Any

logger = logging.getLogger(__name__)

class SolarWindsTxnNameManager:
    """SolarWinds Transaction Name Manager
    Stores trace_id with associate transaction name"""

    def __init__(self, **kwargs: int) -> None:
        self.__cache = {}

    def __delitem__(self, key: str) -> None:
        del self.__cache[key]

    def __getitem__(self, key: str) -> Any:
        return self.__cache[key]

    def __setitem__(self, key: str, value: str) -> None:
        self.__cache[key] = value

    def __str__(self) -> str:
        return "{}".format(self.__cache)

    def get(self, key: str, default: Any = None) -> Any:
        return self.__cache.get(key, default)



""" SolarWinds APM agent-internal logging module.

The solarwinds_apm internal logging is making use of the built-in Python logging module. The solarwinds_apm package
creates a logging handler reporting to stderr where all log messages are reported to. All logging from solarwinds_apm
can be stopped completely by invoking the disable_logger() method.

A log message has the following format:
    '<date> <time> [ <logger_name> <logging_level> p#<process_id>.<thread_id>] <log_message>'

The logging level of the solarwinds_apm logger can be configured with an integer in the range of -1 to 6, representing
the following severities:
    -1: logging disabled
     0: fatal messages
     1: errors
     2: warnings
     3: info
     4: debug low
     5: debug medium
     6: debug high

The log level should not be modified manually from outside the solarwinds_apm package.

For package-internal purposes, the logging level can be configured by providing one of the above levels to the
`set_sw_log_level`. The setLevel method should not be used to modify logging levels of individual module logger
instances.

The name of the main logger is `solarwinds_apm`. Each module should create its own child logger by simply calling
logging.getLogger(__name__) which will return a child logger of the `solarwinds_apm` logger.
"""

import logging
import os


class ApmLoggingLevel:
    """Abstract mapping class providing a conversion between solarwinds_apm agent logging level and Python logging module
    logging levels.
    The solarwinds_apm package has seven different log levels, which are defined in the debug_levels dictionary."""

    # maps string representation of solarwinds_apm.sw_logging levels to their integer counterpart
    debug_levels = {
        'OBOE_DEBUG_DISABLE': -1,
        'OBOE_DEBUG_FATAL': 0,
        'OBOE_DEBUG_ERROR': 1,
        'OBOE_DEBUG_WARNING': 2,
        'OBOE_DEBUG_INFO': 3,
        'OBOE_DEBUG_LOW': 4,
        'OBOE_DEBUG_MEDIUM': 5,
        'OBOE_DEBUG_HIGH': 6,
    }

    # maps solarwinds_apm log levels to python logging module log levels
    logging_map = {
        debug_levels['OBOE_DEBUG_DISABLE']: logging.CRITICAL,
        debug_levels['OBOE_DEBUG_FATAL']: logging.CRITICAL,
        debug_levels['OBOE_DEBUG_ERROR']: logging.ERROR,
        debug_levels['OBOE_DEBUG_WARNING']: logging.WARNING,
        debug_levels['OBOE_DEBUG_INFO']: logging.INFO,
        debug_levels['OBOE_DEBUG_LOW']: logging.DEBUG,
        debug_levels['OBOE_DEBUG_MEDIUM']: logging.DEBUG,
        debug_levels['OBOE_DEBUG_HIGH']: logging.DEBUG,
    }

    @classmethod
    def default_level(cls):
        """Returns integer representation of default debugging level"""
        return cls.debug_levels['OBOE_DEBUG_WARNING']

    @classmethod
    def is_valid_level(cls, level):
        """Returns True if the provided level is a valid interger representation of a solarwinds_apm.sw_logging level,
        False otherwise."""
        try:
            level = int(level)
            return bool(level in list(cls.debug_levels.values()))
        except (ValueError, TypeError):
            return False


def _create_stream_handler():
    """Creates stream handler reporting to stderr."""
    sh = logging.StreamHandler()  # get stream handler to custom logging prefix
    f = logging.Formatter('%(asctime)s [ %(name)s %(levelname)-8s p#%(process)d.%(thread)d] %(message)s')
    sh.setFormatter(f)
    return sh


_stream_handler = _create_stream_handler()


def _get_logger():
    """Creates the logger for agent-internal logging.
    By default, the logging level of the created logger will be set to ApmLoggingLevel.default_level().
    If the logging level of the agent-internal logger needs to be changed, this should happen through one of the
    following options only:
    (1) Through environment variable SW_APM_DEBUG_LEVEL
        - When _get_logger is invoked, SW_APM_DEBUG_LEVEL is checked and the logging level will be set to the
          value provided by the variable. If an invalid value has been provided, the logging level will not be changed.
    (2) By invoking set_sw_log_level
    """

    # create base logger for solarwinds_apm package
    _logger = logging.getLogger('solarwinds_apm')

    # configure logging level of solarwinds_apm logger
    log_level = ApmLoggingLevel.default_level()
    # check if SW_APM_DEBUG_LEVEL has been set and configure newly created logger accordingly
    envv_val = os.getenv('SW_APM_DEBUG_LEVEL', None)
    if envv_val is not None:
        if not ApmLoggingLevel.is_valid_level(envv_val):
            _logger.warning("Misconfigured SW_APM_DEBUG_LEVEL ignored. Defaulted to debug_level %s", log_level)
            _logger.warning("")
        else:
            log_level = int(envv_val)

    _logger.setLevel(ApmLoggingLevel.logging_map[log_level])

    if log_level != ApmLoggingLevel.debug_levels['OBOE_DEBUG_DISABLE']:
        _logger.addHandler(_stream_handler)
    else:
        _logger.propagate = False
        _logger.addHandler(logging.NullHandler())

    return _logger


logger = _get_logger()


def disable_logger(disable=True):
    """Disables all logging messages from the `solarwinds_apm` package when disable is True.
    To restoring logging, set disable as False.
    """
    if disable:
        logger.addHandler(logging.NullHandler())
        logger.removeHandler(_stream_handler)
        logger.propagate = False
    else:
        logger.addHandler(_stream_handler)
        logger.removeHandler(logging.NullHandler())
        logger.propagate = True


def set_sw_log_level(level):
    """Set the logging level of the agent-internal logger to the provided level. This function expects the level
    to be one of the integer representations of the levels defined in ApmLoggingLevel.debug_levels.
    If an invalid level has been provided, the logging level will not be changed but a warning message will be emitted.

    This function should not be used from outside the solarwinds_apm package to modify agent logging behaviour.
    """
    if ApmLoggingLevel.is_valid_level(level):
        logger.setLevel(ApmLoggingLevel.logging_map[level])
        if level == ApmLoggingLevel.debug_levels['OBOE_DEBUG_DISABLE']:
            disable_logger()
    else:
        logger.warning("set_sw_log_level: Ignore attempt to set solarwinds_apm logger to invalid logging level.")



""" This module provides a SolarWinds-specific exporter.

The exporter translates OpenTelemetry spans into SolarWinds Observability events so that the instrumentation data
generated by an OpenTelemetry-based agent can be processed by the SolarWinds backend.
"""

import logging
from typing import Any

from opentelemetry.sdk.trace.export import SpanExporter

from solarwinds_apm.apm_constants import INTL_SWO_SUPPORT_EMAIL
from solarwinds_apm.w3c_transformer import W3CTransformer

logger = logging.getLogger(__name__)


class SolarWindsSpanExporter(SpanExporter):
    """SolarWinds custom span exporter for the SolarWinds backend.
    Initialization requires a liboboe reporter.
    """

    _INTERNAL_TRANSACTION_NAME = "TransactionName"
    _SW_SPAN_KIND = "sw.span_kind"

    def __init__(
        self,
        reporter,
        apm_txname_manager,
        agent_enabled,
        *args,
        **kw_args,
    ) -> None:
        super().__init__(*args, **kw_args)
        self.reporter = reporter
        self.apm_txname_manager = apm_txname_manager
        if agent_enabled:
            from solarwinds_apm.extension.oboe import Context, Metadata
            self.context = Context
            self.metadata = Metadata
        else:
            from solarwinds_apm.apm_noop import Context, Metadata
            self.context = Context
            self.metadata = Metadata

    def export(self, spans) -> None:
        """Export to AO events and report via liboboe.

        Note that OpenTelemetry timestamps are in nanoseconds, whereas SWO expects timestamps
        to be in microseconds, thus all times need to be divided by 1000.
        """
        for span in spans:
            md = self._build_metadata(self.metadata, span.get_span_context())
            if span.parent and span.parent.is_valid:
                # If there is a parent, we need to add an edge to this parent to this entry event
                logger.debug("Continue trace from {}".format(md.toString()))
                parent_md = self._build_metadata(self.metadata, span.parent)
                evt = self.context.createEntry(md, int(span.start_time / 1000),
                                         parent_md)
                if span.parent.is_remote:
                    self._add_info_transaction_name(span, evt)
            else:
                # In OpenTelemrtry, there are no events with individual IDs, but only a span ID
                # and trace ID. Thus, the entry event needs to be generated such that it has the
                # same op ID as the span ID of the OTel span.
                logger.debug("Start a new trace {}".format(md.toString()))
                evt = self.context.createEntry(md, int(span.start_time / 1000))
                self._add_info_transaction_name(span, evt)

            evt.addInfo('Layer', span.name)
            evt.addInfo(self._SW_SPAN_KIND, span.kind.name)
            evt.addInfo('Language', 'Python')
            for k, v in span.attributes.items():
                evt.addInfo(k, v)
            self.reporter.sendReport(evt, False)

            for event in span.events:
                if event.name == 'exception':
                    self._report_exception_event(event)
                else:
                    self._report_info_event(event)

            evt = self.context.createExit(int(span.end_time / 1000))
            evt.addInfo('Layer', span.name)
            self.reporter.sendReport(evt, False)

    def _add_info_transaction_name(self, span, evt) -> None:
        """Add transaction name from cache to root span
        then removes from cache"""
        trace_span_id = "{}-{}".format(span.context.trace_id, span.context.span_id)
        txname = self.apm_txname_manager.get(trace_span_id)
        if txname:
            evt.addInfo(self._INTERNAL_TRANSACTION_NAME, txname)
            del self.apm_txname_manager[trace_span_id]
        else:
            logger.warning(
                "There was an issue setting trace TransactionName. "
                "Please contact {} with this issue".format(INTL_SWO_SUPPORT_EMAIL)
            )

    def _report_exception_event(self, event) -> None:
        evt = self.context.createEvent(int(event.timestamp / 1000))
        evt.addInfo('Label', 'error')
        evt.addInfo('Spec', 'error')
        evt.addInfo('ErrorClass', event.attributes.get('exception.type', None))
        evt.addInfo('ErrorMsg', event.attributes.get('exception.message',
                                                     None))
        evt.addInfo('Backtrace',
                    event.attributes.get('exception.stacktrace', None))
        # add remaining attributes, if any
        for k, v in event.attributes.items():
            if k not in ('exception.type', 'exception.message',
                         'exception.stacktrace'):
                evt.addInfo(k, v)
        self.reporter.sendReport(evt, False)

    def _report_info_event(self, event) -> None:
        print("Found info event")
        print(dir(event))
        print(event)
        evt = self.context.createEvent(int(event.timestamp / 1000))
        evt.addInfo('Label', 'info')
        for k, v in event.attributes.items():
            evt.addInfo(k, v)
        self.reporter.sendReport(evt, False)

    @staticmethod
    def _build_metadata(metadata, span_context) -> Any:
        return metadata.fromString(
            W3CTransformer.traceparent_from_context(span_context)
        )



INTL_SWO_AO_COLLECTOR = "collector.appoptics.com"
INTL_SWO_AO_STG_COLLECTOR = "collector-stg.appoptics.com"
INTL_SWO_COMMA = ","
INTL_SWO_COMMA_W3C_SANITIZED = "...."
INTL_SWO_EQUALS = "="
INTL_SWO_EQUALS_W3C_SANITIZED = "####"
INTL_SWO_TRACESTATE_KEY = "sw"
INTL_SWO_X_OPTIONS_KEY = "sw_xtraceoptions"
INTL_SWO_SIGNATURE_KEY = "sw_signature"
INTL_SWO_DEFAULT_TRACES_EXPORTER = "solarwinds_exporter"
INTL_SWO_TRACECONTEXT_PROPAGATOR = "tracecontext"
INTL_SWO_PROPAGATOR = "solarwinds_propagator"
INTL_SWO_DEFAULT_PROPAGATORS = [
    INTL_SWO_TRACECONTEXT_PROPAGATOR,
    "baggage",
    INTL_SWO_PROPAGATOR,
]
INTL_SWO_DOC_SUPPORTED_PLATFORMS = 'https://documentation.solarwinds.com/en/success_center/observability/default.htm#cshid=app-sysreqs-python-agent'
INTL_SWO_DOC_TRACING_PYTHON = 'https://documentation.solarwinds.com/en/success_center/observability/default.htm#cshid=app-add-python-agent'
INTL_SWO_SUPPORT_EMAIL = 'SWO-support@solarwinds.com'


"""Provides functionality to transform OpenTelemetry Data to SolarWinds Observability data.
"""

import logging
from opentelemetry.sdk.trace import SpanContext

logger = logging.getLogger(__name__)

class W3CTransformer():
    """Transforms inputs to W3C-compliant data for SolarWinds context propagation"""

    _DECISION = "{}"
    _SPAN_ID_HEX = "{:016x}"
    _TRACE_FLAGS_HEX = "{:02x}"
    _TRACE_ID_HEX = "{:032x}"
    _VERSION = "00"

    @classmethod
    def span_id_from_int(cls, span_id: int) -> str:
        """Formats span ID as 16-byte hexadecimal string"""
        return cls._SPAN_ID_HEX.format(span_id)

    @classmethod
    def span_id_from_sw(cls, sw: str) -> str:
        """Formats span ID from sw tracestate value"""
        return sw.split("-")[0]

    @classmethod
    def trace_flags_from_int(cls, trace_flags: int) -> str:
        """Formats trace flags as 8-bit field"""
        return cls._TRACE_FLAGS_HEX.format(trace_flags)

    @classmethod
    def traceparent_from_context(cls, span_context: SpanContext) -> str:
        """Generates a liboboe W3C compatible trace_context from
        provided OTel span context."""
        template = "-".join([
            cls._VERSION,
            cls._TRACE_ID_HEX,
            cls._SPAN_ID_HEX,
            cls._TRACE_FLAGS_HEX
        ])
        xtr = template.format(
            span_context.trace_id,
            span_context.span_id,
            span_context.trace_flags
        )
        logger.debug("Generated traceparent {} from {}".format(xtr, span_context))
        return xtr

    @classmethod
    def sw_from_context(cls, span_context: SpanContext) -> str:
        """Formats tracestate sw value from SpanContext as 16-byte span_id
        with 8-bit trace_flags.
        
        Example: 1a2b3c4d5e6f7g8h-01"""
        sw = "-".join([cls._SPAN_ID_HEX, cls._TRACE_FLAGS_HEX])
        return sw.format(span_context.span_id, span_context.trace_flags)

    @classmethod
    def sw_from_span_and_decision(cls, span_id: int, decision: str) -> str:
        """Formats tracestate sw value from span_id and liboboe decision
        as 16-byte span_id with 8-bit trace_flags.
        
        Example: 1a2b3c4d5e6f7g8h-01"""
        sw = "-".join([cls._SPAN_ID_HEX, cls._DECISION])
        return sw.format(span_id, decision)



"""Module to initialize OpenTelemetry SDK components and liboboe to work with SolarWinds backend"""

import logging
from os import environ
from pkg_resources import (
    iter_entry_points,
    load_entry_point
)
from typing import TYPE_CHECKING

from opentelemetry import trace
from opentelemetry.environment_variables import (
    OTEL_PROPAGATORS,
    OTEL_TRACES_EXPORTER
)
from opentelemetry.instrumentation.propagators import set_global_response_propagator
from opentelemetry.propagate import set_global_textmap
from opentelemetry.propagators.composite import CompositePropagator
from opentelemetry.sdk._configuration import _OTelSDKConfigurator
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor

from solarwinds_apm.apm_constants import (
    INTL_SWO_DEFAULT_PROPAGATORS,
    INTL_SWO_DEFAULT_TRACES_EXPORTER,
    INTL_SWO_SUPPORT_EMAIL,
)
from solarwinds_apm import apm_logging
from solarwinds_apm.apm_config import SolarWindsApmConfig
from solarwinds_apm.apm_txname_manager import SolarWindsTxnNameManager
from solarwinds_apm.response_propagator import SolarWindsTraceResponsePropagator
from solarwinds_apm.inbound_metrics_processor import SolarWindsInboundMetricsSpanProcessor

if TYPE_CHECKING:
    from solarwinds_apm.extension.oboe import Reporter

solarwinds_apm_logger = apm_logging.logger
logger = logging.getLogger(__name__)

class SolarWindsConfigurator(_OTelSDKConfigurator):
    """OpenTelemetry Configurator for initializing APM logger and SolarWinds-reporting SDK components"""

    # Cannot set as env default because not part of OTel Python _KNOWN_SAMPLERS
    # https://github.com/open-telemetry/opentelemetry-python/blob/main/opentelemetry-sdk/src/opentelemetry/sdk/trace/sampling.py#L364-L380
    _DEFAULT_SW_TRACES_SAMPLER = "solarwinds_sampler"

    def _configure(self, **kwargs: int) -> None:
        """Configure SolarWinds APM and OTel components"""
        apm_txname_manager = SolarWindsTxnNameManager()
        apm_config = SolarWindsApmConfig()
        reporter = self._initialize_solarwinds_reporter(apm_config)
        self._configure_otel_components(apm_txname_manager, apm_config, reporter)

    def _configure_otel_components(
        self,
        apm_txname_manager: SolarWindsTxnNameManager,
        apm_config: SolarWindsApmConfig,
        reporter: "Reporter",
    ) -> None:
        """Configure OTel sampler, exporter, propagator, response propagator"""
        self._configure_sampler(apm_config)
        if apm_config.agent_enabled:
            self._configure_metrics_span_processor(
                apm_txname_manager,
                apm_config,
            )
            self._configure_exporter(
                reporter,
                apm_txname_manager,
                apm_config.agent_enabled,
            )
            self._configure_propagator()
            self._configure_response_propagator()
        else:
            # Warning: This may still set OTEL_PROPAGATORS if set because OTel API
            logger.error("Tracing disabled. Not setting propagators.")

    def _configure_sampler(
        self,
        apm_config: SolarWindsApmConfig,
    ) -> None:
        """Always configure SolarWinds OTel sampler, or none if disabled"""
        if not apm_config.agent_enabled:
            logger.error("Tracing disabled. Using OTel no-op tracer provider.")
            trace.set_tracer_provider(
                trace.NoOpTracerProvider()
            )
            return
        try:
            sampler = load_entry_point(
                "solarwinds_apm",
                "opentelemetry_traces_sampler",
                self._DEFAULT_SW_TRACES_SAMPLER
            )(apm_config)
        except:
            logger.exception(
                "Failed to load configured sampler {}. "
                "Please reinstall or contact {}.".format(
                    self._DEFAULT_SW_TRACES_SAMPLER,
                    INTL_SWO_SUPPORT_EMAIL,
                )
            )
            raise
        trace.set_tracer_provider(
            TracerProvider(sampler=sampler)
        )

    def _configure_metrics_span_processor(
        self,
        apm_txname_manager: SolarWindsTxnNameManager,
        apm_config: SolarWindsApmConfig,
    ) -> None:
        """Configure SolarWindsInboundMetricsSpanProcessor"""
        trace.get_tracer_provider().add_span_processor(
            SolarWindsInboundMetricsSpanProcessor(
                apm_txname_manager,
                apm_config.agent_enabled,
            )
        )

    def _configure_exporter(
        self,
        reporter: "Reporter",
        apm_txname_manager: SolarWindsTxnNameManager,
        agent_enabled: bool = True,
    ) -> None:
        """Configure SolarWinds OTel span exporters, defaults or environment
        configured, or none if agent disabled. Initialization of SolarWinds
        exporter requires a liboboe reporter and agent_enabled flag."""
        if not agent_enabled:
            logger.error("Tracing disabled. Cannot set span_processor.")
            return

        # SolarWindsDistro._configure does setdefault so this shouldn't
        # be None, but safer and more explicit this way
        environ_exporter_names = environ.get(
            OTEL_TRACES_EXPORTER,
            INTL_SWO_DEFAULT_TRACES_EXPORTER,
        ).split(",")

        for exporter_name in environ_exporter_names:
            exporter = None
            try:
                if exporter_name == INTL_SWO_DEFAULT_TRACES_EXPORTER:
                    exporter = load_entry_point(
                        "solarwinds_apm",
                        "opentelemetry_traces_exporter",
                        exporter_name
                    )(reporter, apm_txname_manager, agent_enabled)
                else:
                    exporter = next(
                        iter_entry_points(
                            "opentelemetry_traces_exporter",
                            exporter_name
                        )
                    ).load()()
            except:
                # At this point any non-default OTEL_TRACES_EXPORTER has 
                # been checked by ApmConfig so exception here means 
                # something quite wrong
                logger.exception(
                    "Failed to load configured exporter {}. "
                    "Please reinstall or contact {}.".format(
                        exporter_name,
                        INTL_SWO_SUPPORT_EMAIL,
                    )
                )
                raise
            logger.debug("Setting trace with BatchSpanProcessor using {}".format(exporter_name))
            span_processor = BatchSpanProcessor(exporter)
            trace.get_tracer_provider().add_span_processor(span_processor)

    def _configure_propagator(self) -> None:
        """Configure CompositePropagator with SolarWinds and other propagators, default or environment configured"""
        propagators = []

        # SolarWindsDistro._configure does setdefault so this shouldn't
        # be None, but safer and more explicit this way
        environ_propagators_names = environ.get(
            OTEL_PROPAGATORS,
            ",".join(INTL_SWO_DEFAULT_PROPAGATORS),
        ).split(",")

        for propagator_name in environ_propagators_names:
            try:
                propagators.append(
                    next(
                        iter_entry_points("opentelemetry_propagator", propagator_name)
                    ).load()()
                )
            except Exception:
                logger.exception(
                    "Failed to load configured propagator {}".format(
                        propagator_name
                    )
                )
                raise
        logger.debug("Setting CompositePropagator with {}".format(environ_propagators_names))
        set_global_textmap(CompositePropagator(propagators))

    def _configure_response_propagator(self) -> None:
        # Set global HTTP response propagator
        set_global_response_propagator(SolarWindsTraceResponsePropagator())

    def _initialize_solarwinds_reporter(
        self,
        apm_config: SolarWindsApmConfig,
    ) -> "Reporter":
        """Initialize SolarWinds reporter used by sampler and exporter, using SolarWindsApmConfig. This establishes collector and sampling settings in a background thread."""
        if apm_config.agent_enabled:
            from solarwinds_apm.extension.oboe import Reporter
        else:
            from solarwinds_apm.apm_noop import Reporter

        return Reporter(
            hostname_alias=apm_config.get("hostname_alias"),
            log_level=apm_config.get("debug_level"),
            log_file_path=apm_config.get("logname"),
            max_transactions=apm_config.get("max_transactions"),
            max_flush_wait_time=apm_config.get("max_flush_wait_time"),
            events_flush_interval=apm_config.get("events_flush_interval"),
            max_request_size_bytes=apm_config.get("max_request_size_bytes"),
            reporter=apm_config.get("reporter"),
            host=apm_config.get("collector"),
            service_key=apm_config.get("service_key"),
            trusted_path=apm_config.get("trustedpath"),
            buffer_size=apm_config.get("bufsize"),
            trace_metrics=apm_config.get("trace_metrics"),
            histogram_precision=apm_config.get("histogram_precision"),
            token_bucket_capacity=apm_config.get("token_bucket_capacity"),
            token_bucket_rate=apm_config.get("token_bucket_rate"),
            file_single=apm_config.get("reporter_file_single"),
            ec2_metadata_timeout=apm_config.get("ec2_metadata_timeout"),
            grpc_proxy=apm_config.get("proxy"),
            stdout_clear_nonblocking=0,
            is_grpc_clean_hack_enabled=apm_config.get("is_grpc_clean_hack_enabled"),
            w3c_trace_format=1,
            metric_format=apm_config.metric_format,
        )



import logging
import typing

from opentelemetry import trace
from opentelemetry.context.context import Context
from opentelemetry.propagators import textmap
from opentelemetry.trace.span import TraceState

from solarwinds_apm.apm_constants import (
    INTL_SWO_X_OPTIONS_KEY,
    INTL_SWO_SIGNATURE_KEY,
    INTL_SWO_TRACESTATE_KEY
)
from solarwinds_apm.w3c_transformer import W3CTransformer

logger = logging.getLogger(__name__)

class SolarWindsPropagator(textmap.TextMapPropagator):
    """Extracts and injects SolarWinds headers for trace propagation.
    Must be used in composite with TraceContextTextMapPropagator.
    """
    _INVALID_SPAN_ID = 0x0000000000000000
    _TRACESTATE_HEADER_NAME = "tracestate"
    _XTRACEOPTIONS_HEADER_NAME = "x-trace-options"
    _XTRACEOPTIONS_SIGNATURE_HEADER_NAME = "x-trace-options-signature"

    def extract(
        self,
        carrier: textmap.CarrierT,
        context: typing.Optional[Context] = None,
        getter: textmap.Getter = textmap.default_getter,
    ) -> Context:
        """Extracts sw trace options and signature from carrier into OTel
        Context. Note: tracestate is extracted by TraceContextTextMapPropagator
        """
        if context is None:
            context = Context()

        xtraceoptions_header = getter.get(
            carrier,
            self._XTRACEOPTIONS_HEADER_NAME
        )
        if not xtraceoptions_header:
            return context

        context.update({
            INTL_SWO_X_OPTIONS_KEY: xtraceoptions_header[0]
        })
        logger.debug("Extracted {} as {}: {}".format(
            self._XTRACEOPTIONS_HEADER_NAME,
            INTL_SWO_X_OPTIONS_KEY,
            xtraceoptions_header[0]
        ))

        signature_header = getter.get(
            carrier,
            self._XTRACEOPTIONS_SIGNATURE_HEADER_NAME
        )
        if signature_header:
            context.update({
                INTL_SWO_SIGNATURE_KEY: signature_header[0]
            })
            logger.debug("Extracted {} as {}: {}".format(
                self._XTRACEOPTIONS_SIGNATURE_HEADER_NAME,
                INTL_SWO_SIGNATURE_KEY,
                xtraceoptions_header[0]
            ))
        return context

    def inject(
        self,
        carrier: textmap.CarrierT,
        context: typing.Optional[Context] = None,
        setter: textmap.Setter = textmap.default_setter,
    ) -> None:
        """Injects valid sw tracestate from carrier into carrier for HTTP request, to get
        tracestate injected by previous propagators"""
        span = trace.get_current_span(context)
        span_context = span.get_span_context()
        sw_value = W3CTransformer.sw_from_context(span_context)
        trace_state_header = carrier.get(self._TRACESTATE_HEADER_NAME, None)

        # Prepare carrier with carrier's or new tracestate
        trace_state = None
        if not trace_state_header:
            # Only create new trace state if valid span_id
            if span_context.span_id == self._INVALID_SPAN_ID:
                return
            else:
                logger.debug("Creating new trace state for injection with {}".format(sw_value))
                trace_state = TraceState([(INTL_SWO_TRACESTATE_KEY, sw_value)])
        else:
            trace_state = TraceState.from_header([trace_state_header])
            # Check if trace_state already contains sw KV
            if INTL_SWO_TRACESTATE_KEY in trace_state.keys():
                # If so, modify current span_id and trace_flags, and move to beginning of list
                logger.debug("Updating trace state for injection with {}".format(sw_value))
                trace_state = trace_state.update(INTL_SWO_TRACESTATE_KEY, sw_value)

            else:
                # If not, add sw KV to beginning of list
                logger.debug("Adding KV to trace state for injection with {}".format(sw_value))
                trace_state = trace_state.add(INTL_SWO_TRACESTATE_KEY, sw_value)

        setter.set(
            carrier, self._TRACESTATE_HEADER_NAME, trace_state.to_header()
        )

    @property
    def fields(self) -> typing.Set[str]:
        """Returns a set with the fields set in `inject`"""
        return {
            self._TRACESTATE_HEADER_NAME
        }



""" This module provides a SolarWinds-specific sampler.

The custom sampler will fetch sampling configurations for the SolarWinds backend.
"""

import logging
from types import MappingProxyType
from typing import Optional, Sequence, Dict, TYPE_CHECKING

from opentelemetry.context.context import Context as OtelContext
from opentelemetry.sdk.trace.sampling import (Decision, ParentBased, Sampler,
                                              SamplingResult)
from opentelemetry.trace import Link, SpanKind, get_current_span
from opentelemetry.trace.span import SpanContext, TraceState
from opentelemetry.util.types import Attributes

from solarwinds_apm.apm_constants import (
    INTL_SWO_COMMA_W3C_SANITIZED,
    INTL_SWO_EQUALS_W3C_SANITIZED,
    INTL_SWO_TRACESTATE_KEY
)
from solarwinds_apm.apm_config import OboeTracingMode
from solarwinds_apm.traceoptions import XTraceOptions
from solarwinds_apm.w3c_transformer import W3CTransformer

if TYPE_CHECKING:
    from solarwinds_apm.apm_config import SolarWindsApmConfig
    from solarwinds_apm.extension.oboe import Context

logger = logging.getLogger(__name__)


class _SwSampler(Sampler):
    """SolarWinds custom opentelemetry sampler which obeys configuration options provided by the NH/AO Backend."""

    _INTERNAL_BUCKET_CAPACITY = "BucketCapacity"
    _INTERNAL_BUCKET_RATE = "BucketRate"
    _INTERNAL_SAMPLE_RATE = "SampleRate"
    _INTERNAL_SAMPLE_SOURCE = "SampleSource"
    _INTERNAL_SW_KEYS = "SWKeys"
    _LIBOBOE_CONTINUED = -1
    _SW_TRACESTATE_CAPTURE_KEY = "sw.w3c.tracestate"
    _SW_TRACESTATE_ROOT_KEY = "sw.tracestate_parent_id"
    _UNSET = -1
    _XTRACEOPTIONS_RESP_AUTH = "auth"
    _XTRACEOPTIONS_RESP_IGNORED = "ignored"
    _XTRACEOPTIONS_RESP_TRIGGER_IGNORED = "ignored"
    _XTRACEOPTIONS_RESP_TRIGGER_NOT_REQUESTED = "not-requested"
    _XTRACEOPTIONS_RESP_TRIGGER_TRACE = "trigger-trace"

    def __init__(self, apm_config: "SolarWindsApmConfig"):
        self.apm_config = apm_config
        self.context = None
        if self.apm_config.agent_enabled:
            from solarwinds_apm.extension.oboe import Context
            self.context = Context
        else:
            from solarwinds_apm.apm_noop import Context
            self.context = Context

    def get_description(self) -> str:
        return "SolarWinds custom opentelemetry sampler"

    def calculate_liboboe_decision(
        self,
        parent_span_context: SpanContext,
        xtraceoptions: Optional[XTraceOptions] = None,
    ) -> Dict:
        """Calculates oboe trace decision based on parent span context and APM config."""
        tracestring = None
        if parent_span_context.is_valid and parent_span_context.is_remote:
            tracestring = W3CTransformer.traceparent_from_context(
                parent_span_context
            )
        sw_member_value = parent_span_context.trace_state.get(INTL_SWO_TRACESTATE_KEY)

        # 'tracing_mode' is not supported in NH Python, so give as unset
        tracing_mode = self._UNSET

        trigger_trace_mode = OboeTracingMode.get_oboe_trigger_trace_mode(
            self.apm_config.get("trigger_trace")
        )
        # 'sample_rate' is legacy and not supported in NH Python, so give as unset
        sample_rate = self._UNSET

        options = None
        trigger_trace_request = 0
        signature = None
        timestamp = None
        if xtraceoptions:
            options = xtraceoptions.options_header
            trigger_trace_request = xtraceoptions.trigger_trace
            signature = xtraceoptions.signature
            timestamp = xtraceoptions.ts

        logger.debug(
            "Creating new oboe decision with "
            "tracestring: {}, "
            "sw_member_value: {}, "
            "tracing_mode: {}, "
            "sample_rate: {}, "
            "trigger_trace_request: {}, "
            "trigger_trace_mode: {}, "
            "options: {}, "
            "signature: {}, "
            "timestamp: {}".format(
            tracestring,
            sw_member_value,
            tracing_mode,
            sample_rate,
            trigger_trace_request,
            trigger_trace_mode,
            options,
            signature,
            timestamp
        ))
        do_metrics, do_sample, \
            rate, source, bucket_rate, bucket_cap, decision_type, \
            auth, status_msg, auth_msg, status = self.context.getDecisions(
                tracestring,
                sw_member_value,
                tracing_mode,
                sample_rate,
                trigger_trace_request,
                trigger_trace_mode,
                options,
                signature,
                timestamp
            )
        decision = {
            "do_metrics": do_metrics,
            "do_sample": do_sample,
            "rate": rate,
            "source": source,
            "bucket_rate": bucket_rate,
            "bucket_cap": bucket_cap,
            "decision_type": decision_type,
            "auth": auth,
            "status_msg": status_msg,
            "auth_msg": auth_msg,
            "status": status,
        }
        logger.debug("Got liboboe decision outputs: {}".format(decision))
        return decision

    def is_decision_continued(
        self,
        liboboe_decision: Dict
    ) -> bool:
        """Returns True if liboboe decision is a continued one, else False"""
        for val in [
            liboboe_decision["rate"],
            liboboe_decision["source"],
            liboboe_decision["bucket_rate"],
            liboboe_decision["bucket_cap"]
        ]:
            if val != self._LIBOBOE_CONTINUED:
                return False
        return True

    def otel_decision_from_liboboe(
        self,
        liboboe_decision: Dict
    ) -> Decision:
        """Formats OTel decision from liboboe decision"""
        decision = Decision.DROP
        if liboboe_decision["do_sample"]:
            decision = Decision.RECORD_AND_SAMPLE  # even if not do_metrics
        elif liboboe_decision["do_metrics"]:
            decision = Decision.RECORD_ONLY
        logger.debug("OTel decision created: {}".format(decision))
        return decision

    def create_xtraceoptions_response_value(
        self,
        decision: Dict,
        parent_span_context: SpanContext,
        xtraceoptions: XTraceOptions
    ) -> str:
        """Use decision and xtraceoptions to create sanitized xtraceoptions
        response kv with W3C tracestate-allowed delimiters:
        EQUALS_W3C_SANITIZED instead of EQUALS
        COMMA_W3C_SANITIZED instead of COMMA
        """
        response = []

        if xtraceoptions.signature and decision["auth_msg"]:
            response.append(INTL_SWO_EQUALS_W3C_SANITIZED.join([
                self._XTRACEOPTIONS_RESP_AUTH,
                decision["auth_msg"]
            ]))

        if not decision["auth"] or decision["auth"] < 1:
            trigger_msg = ""
            if xtraceoptions.trigger_trace:
                # If a traceparent header was provided then oboe does not generate the message
                tracestring = None
                if parent_span_context.is_valid and parent_span_context.is_remote:
                    tracestring = W3CTransformer.traceparent_from_context(
                        parent_span_context
                    )

                if tracestring and decision["decision_type"] == 0:
                    trigger_msg = self._XTRACEOPTIONS_RESP_TRIGGER_IGNORED
                else:
                    trigger_msg = decision["status_msg"]
            else:
                trigger_msg = self._XTRACEOPTIONS_RESP_TRIGGER_NOT_REQUESTED
            response.append(INTL_SWO_EQUALS_W3C_SANITIZED.join([
                self._XTRACEOPTIONS_RESP_TRIGGER_TRACE,
                trigger_msg
            ]))

        if xtraceoptions.ignored:
            response.append(
                INTL_SWO_EQUALS_W3C_SANITIZED.join([
                    self._XTRACEOPTIONS_RESP_IGNORED,
                    (INTL_SWO_COMMA_W3C_SANITIZED.join(xtraceoptions.ignored))
                ])
            )

        return ";".join(response)

    def create_new_trace_state(
        self,
        decision: Dict,
        parent_span_context: SpanContext,
        xtraceoptions: Optional[XTraceOptions] = None,
    ) -> TraceState:
        """Creates new TraceState based on trace decision, parent span id,
        and x-trace-options if provided"""
        trace_state = TraceState([(
            INTL_SWO_TRACESTATE_KEY,
            W3CTransformer.sw_from_span_and_decision(
                parent_span_context.span_id,
                W3CTransformer.trace_flags_from_int(decision["do_sample"])
            )
        )])
        if xtraceoptions and xtraceoptions.trigger_trace:
            trace_state = trace_state.add(
                XTraceOptions.get_sw_xtraceoptions_response_key(),
                self.create_xtraceoptions_response_value(
                    decision,
                    parent_span_context,
                    xtraceoptions
                )
            )
        logger.debug("Created new trace_state: {}".format(trace_state))
        return trace_state

    def calculate_trace_state(
        self,
        decision: Dict,
        parent_span_context: SpanContext,
        xtraceoptions: Optional[XTraceOptions] = None,
    ) -> TraceState:
        """Calculates trace_state based on parent span context, trace decision,
        and x-trace-options if provided -- for non-existent or remote parent
        spans only."""
        # No valid parent i.e. root span, or parent is remote
        if not parent_span_context.is_valid:
            trace_state = self.create_new_trace_state(
                decision,
                parent_span_context,
                xtraceoptions
            )
        else:
            trace_state = parent_span_context.trace_state
            if not trace_state:
                # tracestate nonexistent/non-parsable
                trace_state = self.create_new_trace_state(
                    decision,
                    parent_span_context,
                    xtraceoptions
                )
            else:
                # Update trace_state with span_id and sw trace_flags
                trace_state = trace_state.update(
                    INTL_SWO_TRACESTATE_KEY,
                    W3CTransformer.sw_from_span_and_decision(
                        parent_span_context.span_id,
                        W3CTransformer.trace_flags_from_int(decision["do_sample"])
                    )
                )
                # Update trace_state with x-trace-options-response
                # Not a propagated header, so always an add
                if xtraceoptions and xtraceoptions.trigger_trace:
                    trace_state = trace_state.add(
                        XTraceOptions.get_sw_xtraceoptions_response_key(),
                        self.create_xtraceoptions_response_value(
                            decision,
                            parent_span_context,
                            xtraceoptions
                        )
                    )
                logger.debug("Updated trace_state: {}".format(trace_state))
        return trace_state

    def remove_response_from_sw(
        self,
        trace_state: TraceState
    ) -> TraceState:
        """Remove xtraceoptions response from tracestate"""
        return trace_state.delete(XTraceOptions.get_sw_xtraceoptions_response_key())

    def add_tracestate_capture_to_attributes_dict(
        self,
        attributes_dict: dict,
        decision: Dict,
        trace_state: TraceState,
        parent_span_context: SpanContext,
    ) -> dict:
        """Calculate and add SW tracestate capture to attributes_dict,
        which is a dict not an Attributes/MappingProxy object"""
        tracestate_capture = attributes_dict.get(self._SW_TRACESTATE_CAPTURE_KEY, None)
        if not tracestate_capture:
            trace_state_no_response = self.remove_response_from_sw(trace_state)
        else:
            # Must retain all potential tracestate pairs for attributes
            attr_trace_state = TraceState.from_header([
                tracestate_capture
            ])
            new_attr_trace_state = attr_trace_state.update(
                INTL_SWO_TRACESTATE_KEY,
                W3CTransformer.sw_from_span_and_decision(
                    parent_span_context.span_id,
                    W3CTransformer.trace_flags_from_int(decision["do_sample"])
                )
            )
            trace_state_no_response = self.remove_response_from_sw(new_attr_trace_state)
        attributes_dict[self._SW_TRACESTATE_CAPTURE_KEY] = trace_state_no_response.to_header()
        return attributes_dict

    def calculate_attributes(
        self,
        span_name: str,
        attributes: Attributes,
        decision: Dict,
        trace_state: TraceState,
        parent_span_context: SpanContext,
        xtraceoptions: XTraceOptions
    ) -> Attributes or None:
        """Calculates Attributes or None based on trace decision, trace state,
        parent span context, xtraceoptions, and existing attributes."""
        logger.debug("Received attributes: {}".format(attributes))
        # Don't set attributes if not tracing
        otel_decision = self.otel_decision_from_liboboe(decision)
        if not Decision.is_sampled(otel_decision):
            logger.debug("Trace decision not is_sampled - not setting attributes")
            return None
        
        new_attributes = {}

        # Always (root or is_remote) set _INTERNAL_SW_KEYS if injected
        internal_sw_keys = xtraceoptions.sw_keys
        if internal_sw_keys:
            new_attributes[self._INTERNAL_SW_KEYS] = internal_sw_keys
    
        # Always (root or is_remote) set service entry internal KVs       
        new_attributes[self._INTERNAL_BUCKET_CAPACITY] = "{}".format(decision["bucket_cap"])
        new_attributes[self._INTERNAL_BUCKET_RATE] = "{}".format(decision["bucket_rate"])
        new_attributes[self._INTERNAL_SAMPLE_RATE] = decision["rate"]
        new_attributes[self._INTERNAL_SAMPLE_SOURCE] = decision["source"]
        logger.debug(
            "Set attributes with service entry internal KVs: {}".format(new_attributes)
        )

        # Trace's root span has no valid traceparent nor tracestate
        # so we can't calculate remaining attributes
        if not parent_span_context.is_valid or not trace_state:
            logger.debug(
                "No valid traceparent or no tracestate - returning attributes: {}"
                .format(new_attributes)
            )

            if new_attributes:
                # attributes must be immutable for SamplingResult
                return MappingProxyType(new_attributes)
            else:
                return None

        if not attributes:
            # _SW_TRACESTATE_ROOT_KEY is set once per trace, if possible
            sw_value = parent_span_context.trace_state.get(INTL_SWO_TRACESTATE_KEY, None)
            if sw_value:
                new_attributes[self._SW_TRACESTATE_ROOT_KEY]= W3CTransformer.span_id_from_sw(sw_value)
        else:
            # Copy existing MappingProxyType KV into new_attributes for modification.
            # attributes may have other vendor info etc
            for k,v in attributes.items():
                new_attributes[k] = v

        new_attributes = self.add_tracestate_capture_to_attributes_dict(
            new_attributes,
            decision,
            trace_state,
            parent_span_context,
        )

        logger.debug("Setting attributes: {}".format(new_attributes))

        # attributes must be immutable for SamplingResult
        return MappingProxyType(new_attributes)

    def should_sample(
        self,
        parent_context: Optional[OtelContext],
        trace_id: int,
        name: str,
        kind: SpanKind = None,
        attributes: Attributes = None,
        links: Sequence["Link"] = None,
        trace_state: "TraceState" = None,
    ) -> "SamplingResult":
        """
        Calculates SamplingResult based on calculation of new/continued trace
        decision, new/updated trace state, and new/updated attributes.
        """
        parent_span_context = get_current_span(
            parent_context
        ).get_span_context()
        xtraceoptions = XTraceOptions(parent_context)

        liboboe_decision = self.calculate_liboboe_decision(
            parent_span_context,
            xtraceoptions
        )

        # Always calculate trace_state for propagation
        new_trace_state = self.calculate_trace_state(
            liboboe_decision,
            parent_span_context,
            xtraceoptions
        )
        new_attributes = self.calculate_attributes(
            name,
            attributes,
            liboboe_decision,
            new_trace_state,
            parent_span_context,
            xtraceoptions
        )
        otel_decision = self.otel_decision_from_liboboe(liboboe_decision)

        return SamplingResult(
            otel_decision,
            new_attributes if Decision.is_sampled(otel_decision) else None,
            new_trace_state,
        )


class ParentBasedSwSampler(ParentBased):
    """
    Sampler that respects its parent span's sampling decision, but otherwise
    samples according to the configurations from the NH/AO backend.

    Requires "SolarWindsApmConfig".
    """
    def __init__(self, apm_config: "SolarWindsApmConfig"):
        """
        Uses _SwSampler/liboboe if no parent span.
        Uses _SwSampler/liboboe if parent span is_remote.
        Uses OTEL defaults if parent span is_local.
        """
        super().__init__(
            root=_SwSampler(apm_config),
            remote_parent_sampled=_SwSampler(apm_config),
            remote_parent_not_sampled=_SwSampler(apm_config)
        )

    def should_sample(
        self,
        parent_context: Optional["Context"],
        trace_id: int,
        name: str,
        kind: SpanKind = None,
        attributes: Attributes = None,
        links: Sequence["Link"] = None,
        trace_state: "TraceState" = None
    ) -> "SamplingResult":

        return super().should_sample(
            parent_context,
            trace_id,
            name,
            kind,
            attributes,
            links,
            trace_state
        )



import logging
import re
import typing

from opentelemetry.context.context import Context

from solarwinds_apm.apm_constants import (
    INTL_SWO_X_OPTIONS_KEY,
    INTL_SWO_SIGNATURE_KEY
)

logger = logging.getLogger(__name__)

class XTraceOptions():
    """Formats X-Trace-Options and signature for trigger tracing"""

    _SW_XTRACEOPTIONS_RESPONSE_KEY = "xtrace_options_response"
    _XTRACEOPTIONS_CUSTOM = (r"^custom-[^\s]*$")
    _XTRACEOPTIONS_CUSTOM_RE = re.compile(_XTRACEOPTIONS_CUSTOM)

    _XTRACEOPTIONS_HEADER_KEY_SW_KEYS = "sw-keys"
    _XTRACEOPTIONS_HEADER_KEY_TRIGGER_TRACE = "trigger-trace"
    _XTRACEOPTIONS_HEADER_KEY_TS = "ts"

    def __init__(self,
        context: typing.Optional[Context] = None,
    ):
        """
        Args:
          context: OTel context that may contain OTEL_CONTEXT_SW_OPTIONS_KEY,OTEL_CONTEXT_SW_SIGNATURE_KEY
        """
        self.ignored = []
        self.options_header = ""
        self.signature = None
        self.sw_keys = ""
        self.trigger_trace = 0
        self.ts = 0
        
        if not context:
            return
        options_header = context.get(INTL_SWO_X_OPTIONS_KEY, None)
        if not options_header:
            return

        # store original header for sample decision later
        self.options_header = options_header

        traceoptions = re.split(r";+", options_header)
        for option in traceoptions:
            # KVs (e.g. sw-keys or custom-key1) are assigned by equals
            option_kv = option.split("=", 2)
            if not option_kv[0]:
                continue

            option_key = option_kv[0].strip()
            if option_key == self._XTRACEOPTIONS_HEADER_KEY_TRIGGER_TRACE:
                if len(option_kv) > 1:
                    logger.debug("trigger-trace must be standalone flag. Ignoring.")
                    self.ignored.append(
                        self._XTRACEOPTIONS_HEADER_KEY_TRIGGER_TRACE
                    )
                else:
                    self.trigger_trace = 1
        
            elif option_key == self._XTRACEOPTIONS_HEADER_KEY_SW_KEYS:
                self.sw_keys = option_kv[1].strip()                    

            elif re.match(self._XTRACEOPTIONS_CUSTOM_RE, option_key):
                # custom keys are valid but do not need parsing
                pass

            elif option_key == self._XTRACEOPTIONS_HEADER_KEY_TS:
                try:
                    self.ts = int(option_kv[1])
                except ValueError as e:
                    logger.debug("ts must be base 10 int. Ignoring.")
                    self.ignore.append(self._XTRACEOPTIONS_HEADER_KEY_TS)
            
            else:
                logger.debug(
                    "{} is not a recognized trace option. Ignoring".format(
                        option_key
                    ))
                self.ignored.append(option_key)

            if self.ignored:
                logger.debug(
                    "Some x-trace-options were ignored: {}".format(
                        ", ".join(self.ignored)
                    ))
        
        options_signature = context.get(INTL_SWO_SIGNATURE_KEY, None)
        if options_signature:
            self.signature = options_signature

    @classmethod
    def get_sw_xtraceoptions_response_key(cls) -> str:
        return cls._SW_XTRACEOPTIONS_RESPONSE_KEY



#!/usr/bin/env python

#    Copyright 2022 SolarWinds, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable-msg=missing-module-docstring
import logging
import os
import sys

from setuptools import (
    Extension,
    setup
)
from setuptools.command.build_ext import build_ext
from setuptools.command.build_py import build_py


logger = logging.getLogger(__name__)

def is_alpine_distro():
    """Checks if current system is Alpine Linux."""
    if os.path.exists("/etc/alpine-release"):
        return True

    try:
        with open("/etc/os-release", 'r') as osr:
            releases = osr.readlines()
            releases = [l[:-1] for l in releases]
        if 'NAME="Alpine Linux"' in releases:
            return True
    except Exception:
        pass

    return False

def python_version_supported():
    if sys.version_info[0] == 3 and sys.version_info[1] >= 6:
        return True
    return False

def os_supported():
    return sys.platform.startswith('linux')

def link_oboe_lib(src_lib):
    """Set up the C-extension libraries.

    Create two .so library symlinks, namely 'liboboe-1.0.so' and 'liboboe-1.0.so.0 which are needed when the
    solarwinds_apm package is built from source. This step is needed since Oboe library is platform specific.

    The src_lib parameter is the name of the library file under solarwinds_apm/extension the above mentioned symlinks will
    point to. If a file with the provided name does not exist, no symlinks will be created."""

    logger.info("Create links to platform specific liboboe library file")
    link_dst = ('liboboe-1.0.so', 'liboboe-1.0.so.0')
    cwd = os.getcwd()
    try:
        os.chdir('./solarwinds_apm/extension/')
        if not os.path.exists(src_lib):
            raise Exception("C-extension library file {} does not exist.".format(src_lib))
        for dst in link_dst:
            if os.path.exists(dst):
                # if the destination library files exist already, they need to be deleted, otherwise linking will fail
                os.remove(dst)
                logger.info("Removed %s" % dst)
            os.symlink(src_lib, dst)
            logger.info("Created new link at {} to {}".format(dst, src_lib))
    except Exception as ecp:
        logger.info("[SETUP] failed to set up links to C-extension library: {e}".format(e=ecp))
    finally:
        os.chdir(cwd)

class CustomBuild(build_py):
    def run(self):
        self.run_command('build_ext')
        build_py.run(self)

class CustomBuildExt(build_ext):
    def run(self):
        if sys.platform == 'darwin':
            return

        oboe_lib = "liboboe-1.0-alpine-x86_64.so.0.0.0" if is_alpine_distro() else "liboboe-1.0-x86_64.so.0.0.0"
        link_oboe_lib(oboe_lib)
        build_ext.run(self)

class CustomBuildExtLambda(build_ext):
    def run(self):
        link_oboe_lib("liboboe-1.0-lambda-x86_64.so.0.0.0")
        build_ext.run(self)


if not (python_version_supported() and os_supported()):
    logger.warn(
        "[SETUP] This package supports only Python 3.7 and above on Linux. "
        "Other platform or python versions may not work as expected.")

ext_modules = [
    Extension('solarwinds_apm.extension._oboe',
              sources=[
                  'solarwinds_apm/extension/oboe_wrap.cxx',
                  'solarwinds_apm/extension/oboe_api.cpp'
              ],
              depends=[
                  'solarwinds_apm/extension/oboe_api.hpp',
              ],
              include_dirs=[
                  'solarwinds_apm/extension',
                  'solarwinds_apm'
              ],
              libraries=['oboe-1.0', 'rt'],
              library_dirs=['solarwinds_apm/extension'],
              extra_compile_args=["-std=c++11"],
              runtime_library_dirs=['$ORIGIN']),
]

setup(
    cmdclass={
        'build_ext': CustomBuildExt,
        'build_ext_lambda': CustomBuildExtLambda,
        'build_py': CustomBuild,
    },
    ext_modules=ext_modules,
)




from collections import defaultdict
from functools import reduce
import logging
import os
from pkg_resources import iter_entry_points
import sys
from typing import Any

from opentelemetry.environment_variables import (
    OTEL_PROPAGATORS,
    OTEL_TRACES_EXPORTER,
)

from solarwinds_apm import apm_logging
from solarwinds_apm.apm_constants import (
    INTL_SWO_AO_COLLECTOR,
    INTL_SWO_AO_STG_COLLECTOR,
    INTL_SWO_DEFAULT_TRACES_EXPORTER,
    INTL_SWO_DEFAULT_PROPAGATORS,
    INTL_SWO_DOC_SUPPORTED_PLATFORMS,
    INTL_SWO_DOC_TRACING_PYTHON,
    INTL_SWO_SUPPORT_EMAIL,
    INTL_SWO_PROPAGATOR,
    INTL_SWO_TRACECONTEXT_PROPAGATOR,
)

logger = logging.getLogger(__name__)

class OboeTracingMode:
    """Provides an interface to translate the string representation of tracing_mode to the C-extension equivalent."""
    OBOE_SETTINGS_UNSET = -1
    OBOE_TRACE_DISABLED = 0
    OBOE_TRACE_ENABLED = 1
    OBOE_TRIGGER_DISABLED = 0
    OBOE_TRIGGER_ENABLED = 1

    @classmethod
    def get_oboe_trace_mode(cls, tracing_mode: str) -> int:
        if tracing_mode == 'enabled':
            return cls.OBOE_TRACE_ENABLED
        if tracing_mode == 'disabled':
            return cls.OBOE_TRACE_DISABLED
        return cls.OBOE_SETTINGS_UNSET

    @classmethod
    def get_oboe_trigger_trace_mode(cls, trigger_trace_mode: str) -> int:
        if trigger_trace_mode == 'enabled':
            return cls.OBOE_TRIGGER_ENABLED
        if trigger_trace_mode == 'disabled':
            return cls.OBOE_TRIGGER_DISABLED
        return cls.OBOE_SETTINGS_UNSET


class SolarWindsApmConfig:
    """SolarWinds APM Configuration Class
    The precedence: in-code keyword arguments > Environment Variables > config file > default values.
    Note that oboe doesn't read configurations by itself. The Python agent needs to
    read environment variables and/or config files and pass them into oboe. This is
    done only once during the initialization and the properties cannot be refreshed.
    """
    _DELIMITER = '.'
    _KEY_MASK = '{}...{}:{}'
    _KEY_MASK_BAD_FORMAT = '{}...<invalid_format>'
    _KEY_MASK_BAD_FORMAT_SHORT = '{}<invalid_format>'

    def __init__(self, **kwargs: int) -> None:
        self.__config = dict()
        # Update the config with default values
        self.__config = {
            # 'tracing_mode' is unset by default and not supported in NH Python
            'tracing_mode': None,
            # 'trigger_trace' is enabled by default
            'trigger_trace': 'enabled',
            'collector': '',  # the collector address in host:port format.
            'reporter': '',  # the reporter mode, either 'udp' or 'ssl'.
            'debug_level': apm_logging.ApmLoggingLevel.default_level(),
            'service_key': '',
            'hostname_alias': '',
            'trustedpath': '',
            'events_flush_interval': -1,
            'max_request_size_bytes': -1,
            'ec2_metadata_timeout': 1000,
            'max_flush_wait_time': -1,
            'max_transactions': -1,
            'logname': '',
            'trace_metrics': -1,
            'token_bucket_capacity': -1,
            'token_bucket_rate': -1,
            'bufsize': -1,
            'histogram_precision': -1,
            'reporter_file_single': 0,
            'enable_sanitize_sql': True,
            'inst_enabled': defaultdict(lambda: True),
            'log_trace_id': 'never',
            'proxy': '',
            'transaction': defaultdict(lambda: True),
            'inst': defaultdict(lambda: True),
            'is_grpc_clean_hack_enabled': False,
        }
        self.agent_enabled = self._calculate_agent_enabled()

        if self.agent_enabled:
            from solarwinds_apm.extension.oboe import Context
            self.context = Context
        else:
            from solarwinds_apm.apm_noop import Context
            self.context = Context

        # TODO Implement config with cnf_file after alpha
        # cnf_file = os.environ.get('SW_APM_APM_CONFIG_PYTHON', os.environ.get('SW_APM_PYCONF', None))
        # if cnf_file:
        #     self.update_with_cnf_file(cnf_file)

        self.update_with_env_var()

        self.metric_format = self._calculate_metric_format()

        # TODO Implement in-code config with kwargs after alpha
        # self.update_with_kwargs(kwargs)

        logger.debug("Set ApmConfig as: {}".format(self))

    def _is_lambda(self) -> bool:
        """Checks if agent is running in an AWS Lambda environment."""
        if os.environ.get('AWS_LAMBDA_FUNCTION_NAME') and os.environ.get("LAMBDA_TASK_ROOT"):
            logger.warning("AWS Lambda is not yet supported by Python SolarWinds APM.")
            return True
        return False

    def _calculate_agent_enabled(self) -> bool:
        """Checks if agent is enabled/disabled based on config:
        - SW_APM_SERVICE_KEY   (required)
        - OTEL_PROPAGATORS     (optional)
        - OTEL_TRACES_EXPORTER (optional)
        - SW_APM_AGENT_ENABLED (optional)
        """
        agent_enabled = True

        try:
            # SolarWindsDistro._configure does setdefault so this shouldn't
            # be None, but safer and more explicit this way
            environ_propagators = os.environ.get(
                OTEL_PROPAGATORS,
                ",".join(INTL_SWO_DEFAULT_PROPAGATORS),
            ).split(",")
            # If not using the default propagators,
            # can any arbitrary list BUT
            # (1) must include tracecontext and solarwinds_propagator
            # (2) tracecontext must be before solarwinds_propagator
            if environ_propagators != INTL_SWO_DEFAULT_PROPAGATORS:
                if not INTL_SWO_TRACECONTEXT_PROPAGATOR in environ_propagators or \
                    not INTL_SWO_PROPAGATOR in environ_propagators:
                    logger.error("Must include tracecontext and solarwinds_propagator in OTEL_PROPAGATORS to use SolarWinds APM. Tracing disabled.")
                    raise ValueError

                if environ_propagators.index(INTL_SWO_PROPAGATOR) \
                    < environ_propagators.index(INTL_SWO_TRACECONTEXT_PROPAGATOR):
                    logger.error("tracecontext must be before solarwinds_propagator in OTEL_PROPAGATORS to use SolarWinds APM. Tracing disabled.")
                    raise ValueError
        except ValueError:
            agent_enabled = False

        try:
            # SolarWindsDistro._configure does setdefault so this shouldn't
            # be None, but safer and more explicit this way
            environ_exporters = os.environ.get(
                OTEL_TRACES_EXPORTER,
                INTL_SWO_DEFAULT_TRACES_EXPORTER,
            ).split(",")
            # If not using the default propagators,
            # can any arbitrary list BUT
            # (1) must include solarwinds_exporter
            # (2) other exporters must be loadable by OTel
            if INTL_SWO_DEFAULT_TRACES_EXPORTER not in environ_exporters:
                logger.error("Must include solarwinds_exporter in OTEL_TRACES_EXPORTER to use Solarwinds APM. Tracing disabled.")
                raise ValueError
            for environ_exporter_name in environ_exporters:
                try:
                    if environ_exporter_name != INTL_SWO_DEFAULT_TRACES_EXPORTER:
                        next(
                            iter_entry_points(
                                "opentelemetry_traces_exporter",
                                environ_exporter_name
                            )
                        )
                except StopIteration:
                    logger.error(
                        "Failed to load configured OTEL_TRACES_EXPORTER {}. "
                        "Tracing disabled".format(
                            environ_exporter_name
                        )
                    )
                    agent_enabled = False
        except ValueError:
            agent_enabled = False

        try:
            if os.environ.get('SW_APM_AGENT_ENABLED', 'true').lower() == 'false':
                agent_enabled = False
                logger.info(
                    "SolarWinds APM is disabled and will not report any traces because the environment variable "
                    "SW_APM_AGENT_ENABLED is set to 'false'! If this is not intended either unset the variable or set it to "
                    "a value other than false. Note that the value of SW_APM_AGENT_ENABLED is case-insensitive.")
                raise ImportError

            if not os.environ.get('SW_APM_SERVICE_KEY', None) and not self._is_lambda():
                logger.error("Missing service key. Tracing disabled.")
                agent_enabled = False
                raise ImportError

            else:
                # Key must be at least one char + ":" + at least one other char
                key_parts = [p for p in os.environ.get('SW_APM_SERVICE_KEY', "").split(":") if len(p) > 0]
                if len(key_parts) != 2:
                    logger.error("Incorrect service key format. Tracing disabled.")
                    agent_enabled = False
                    raise ImportError

        except ImportError as e:
            try:
                if agent_enabled:
                    # only log the following messages if agent wasn't explicitly disabled
                    # via SW_APM_AGENT_ENABLED or due to missing service key
                    if sys.platform.startswith('linux'):
                        logger.warning(
                            """Missing extension library.
                            Tracing is disabled and will go into no-op mode.
                            Contact {} if this is unexpected.
                            Error: {}
                            See: {}""".format(
                                INTL_SWO_SUPPORT_EMAIL,
                                e,
                                INTL_SWO_DOC_TRACING_PYTHON,
                            ))
                    else:
                        logger.warning(
                            """Platform {} not yet supported.
                            See: {}
                            Tracing is disabled and will go into no-op mode.
                            Contact {} if this is unexpected.""".format(
                                sys.platform,
                                INTL_SWO_DOC_SUPPORTED_PLATFORMS,
                                INTL_SWO_SUPPORT_EMAIL,
                            ))
            except ImportError as err:
                logger.error(
                    """Unexpected error: {}.
                    Please reinstall or contact {}.""".format(
                        err,
                        INTL_SWO_SUPPORT_EMAIL,
                    ))
            finally:
                # regardless of how we got into this (outer) exception block, the agent will not be able to trace (and thus be
                # disabled)
                agent_enabled = False
        
        logger.debug("agent_enabled: {}".format(agent_enabled))
        return agent_enabled

    def _calculate_metric_format(self) -> int:
        """Return one of 0 (both) or 1 (TransactionResponseTime only). Future: return 2 (ResponseTime only) instead of 0. Based on collector URL which may have a port e.g. foo-collector.com:443"""
        metric_format = 0
        host = self.get("collector")
        if host:
            if len(host.split(":")) > 1:
                host = host.split(":")[0]
            if host in [INTL_SWO_AO_COLLECTOR, INTL_SWO_AO_STG_COLLECTOR]:
                logger.warning("AO collector detected. Only exporting TransactionResponseTime metrics")
                metric_format = 1
        return metric_format

    def mask_service_key(self) -> str:
        """Return masked service key except first 4 and last 4 chars"""
        service_key = self.__config.get('service_key')
        if not service_key:
            return ""
        if service_key.strip() == "":
            return service_key

        key_parts = service_key.split(":")
        if len(key_parts) < 2:
            bad_format_key = key_parts[0]
            if len(bad_format_key) < 5:
                return self._KEY_MASK_BAD_FORMAT_SHORT.format(bad_format_key)
            return self._KEY_MASK_BAD_FORMAT.format(
                bad_format_key[0:4],
            )

        api_token = key_parts[0]
        if len(api_token) < 9:
            return service_key
        return self._KEY_MASK.format(
            api_token[0:4],
            api_token[-4:],
            key_parts[1],
        )

    def _config_mask_service_key(self) -> dict:
        """Return new config with service key masked"""
        config_masked = {}
        for k, v in self.__config.items():
            if k == "service_key":
                v = self.mask_service_key()
            config_masked[k] = v
        return config_masked

    def __str__(self) -> str:
        """String representation of ApmConfig is config with masked service key,
        plus agent_enabled and context"""
        apm_config = {
            "__config": self._config_mask_service_key(),
            "agent_enabled": self.agent_enabled,
            "context": self.context,
        }
        return "{}".format(apm_config)

    def __setitem__(self, key: str, value: str) -> None:
        """Refresh the configurations in liboboe global struct while user changes settings.
        """
        if key == 'tracing_mode':
            self._set_config_value(key, value)

        # TODO set up using OTel logging instrumentation
        elif key == 'log_trace_id':
            self._set_config_value(key, value)

        elif key in ('enable_sanitize_sql', 'warn_deprecated'):
            self._set_config_value(key, value)
        else:
            logger.warning('Unsupported SolarWinds APM config key: {key}'.format(key=key))

    def __getitem__(self, key: str) -> Any:
        return self.__config[key]

    def __delitem__(self, key: str) -> None:
        del self.__config[key]

    def get(self, key: str, default: Any = None):
        """Get the value of key. Nested keys separated by a dot are also accepted.
        Suggestion: Use mask_service_key() to safely get service_key value"""
        key = key.split(self._DELIMITER)
        value = reduce(lambda d, k: d.get(k, None) if isinstance(d, dict) else None, key, self.__config)
        return value if value is not None else default

    def update_with_cnf_file(self, cnf_path: str) -> None:
        """Update the settings with the config file, if any."""
        # TODO Implement config with cnf_file after alpha
        pass

    def update_with_env_var(self) -> None:
        """Update the settings with environment variables."""
        val = os.environ.get('SW_APM_PREPEND_DOMAIN_NAME', None)
        if val is not None:
            self._set_config_value('transaction.prepend_domain_name', val)
        available_envvs = set(self.__config.keys())
        # TODO after alpha: is_lambda
        for key in available_envvs:
            if key in ('inst_enabled', 'transaction', 'inst'):
                # we do not allow complex config options to be set via environment variables
                continue
            env = 'SW_APM_' + key.upper()
            val = os.environ.get(env)
            if val is not None:
                self._set_config_value(key, val)

    def update_with_kwargs(self, kwargs):
        """Update the configuration settings with (in-code) keyword arguments"""
        # TODO Implement in-code config with kwargs after alpha
        pass

    def _set_config_value(self, keys: str, val: Any) -> Any:
        """Sets the value of the config option indexed by 'keys' to 'val', where 'keys' is a nested key (separated by
        self.delimiter, i.e., the position of the element to be changed in the nested dictionary)"""
        def _convert_to_bool(val):
            """Converts given value to boolean value"""
            val = val.lower() if isinstance(val, str) else val
            return val == 'true' if isinstance(val, str) and val in ('true', 'false') else bool(int(val))

        # _config is a nested dict, thus find most deeply nested sub dict according to the provided keys
        # by defaulting to None in d.get(), we do not allow the creation of any new (key, value) pair, even
        # when we are handling a defaultdict (i.e., with this we do not allow e.g. the creation of new instrumentations
        # through the config)
        keys = keys.split(self._DELIMITER)
        sub_dict = reduce(lambda d, key: d.get(key, None) if isinstance(d, dict) else None, keys[:-1], self.__config)
        key = keys[-1]
        try:
            if keys == ['ec2_metadata_timeout']:
                timeout = int(val)
                if timeout not in range(0, 3001):
                    raise ValueError
                self.__config[key] = timeout
            elif keys == ['token_bucket_capacity']:
                bucket_cap = float(val)
                if not 0 <= bucket_cap <= 8.0:
                    raise ValueError
                self.__config[key] = bucket_cap
            elif keys == ['token_bucket_rate']:
                bucket_rate = float(val)
                if not 0 <= bucket_rate <= 4.0:
                    raise ValueError
                self.__config[key] = bucket_rate
            elif keys == ['proxy']:
                if not isinstance(val, str) or not val.startswith('http://'):
                    raise ValueError
                self.__config[key] = val
            elif keys == ['tracing_mode']:
                if not isinstance(val, str):
                    raise ValueError
                val = val.lower()
                if val in ['always', 'never']:
                    val = 'enabled' if val == 'always' else 'disabled'
                if val not in ['enabled', 'disabled']:
                    raise ValueError
                self.__config[key] = val
                self.context.setTracingMode(OboeTracingMode.get_oboe_trace_mode(val))
            elif keys == ['trigger_trace']:
                if not isinstance(val, str):
                    raise ValueError
                val = val.lower()
                if val in ['always', 'never']:
                    val = 'enabled' if val == 'always' else 'disabled'
                if val not in ['enabled', 'disabled']:
                    raise ValueError
                self.__config[key] = val
                self.context.setTriggerMode(OboeTracingMode.get_oboe_trigger_trace_mode(val))
            elif keys == ['reporter']:
                # TODO: support 'lambda'
                if not isinstance(val, str) or val.lower() not in ('udp', 'ssl', 'null', 'file'):
                    raise ValueError
                self.__config[key] = val.lower()
            elif keys == ['debug_level']:
                val = int(val)
                if not apm_logging.ApmLoggingLevel.is_valid_level(val):
                    raise ValueError
                self.__config[key] = val
                # update logging level of agent logger
                apm_logging.set_sw_log_level(val)
            elif keys == ['log_trace_id']:
                if not isinstance(val, str) or val.lower() not in [
                        'never',
                        'sampled',
                        'traced',
                        'always',
                ]:
                    raise ValueError
                self.__config[key] = val.lower()
            elif keys == ['is_grpc_clean_hack_enabled']:
                self.__config[key] = _convert_to_bool(val)
            elif isinstance(sub_dict, dict) and keys[-1] in sub_dict:
                if isinstance(sub_dict[keys[-1]], bool):
                    val = _convert_to_bool(val)
                else:
                    val = type(sub_dict[keys[-1]])(val)
                sub_dict[keys[-1]] = val
            else:
                logger.warning("Ignore invalid configuration key: {}".format('.'.join(keys)))
        except (ValueError, TypeError):
            logger.warning(
                'Ignore config option with invalid (non-convertible or out-of-range) type: ' +
                '.'.join(keys if keys[0] not in ['inst', 'transaction'] else keys[1:]))
